from __future__ import absolute_import

import csv
import io
import re

from .base import HttpSite


class CsvSite(HttpSite):
    _delim = None

    @property
    def dialect(self):
        if "pattern" not in self.conf:
            return "excel"

        class DelimDialect(csv.excel):
            delimiter = str(self.delim)
            skipinitialspace = True

        return DelimDialect()

    @property
    def delim(self):
        return self._delim or self.conf.get("pattern", ",")

    def get_content(self):
        r = super(CsvSite, self).get_content()
        body = r.text

        if len(self.delim) > 1:
            body = re.sub(self.conf["pattern"], "|", body)
            self._delim = "|"

        buf = io.StringIO(body)
        csvfile = csv.reader(buf, dialect=self.dialect)

        return csvfile

    def run(self):
        r = self._req(self.conf["request"])

        body = r.text
        if len(self.delim) > 1:
            body = re.sub(self.conf["pattern"], "|", body)
            self._delim = "|"

        buf = io.StringIO(body)
        csvfile = csv.reader(buf, dialect=self.dialect)

        for (lineno, row) in enumerate(csvfile):
            for parser in self.conf["results"]:
                start = parser.get("start", 1)
                stop = parser.get("end", None)

                # raise ValueError(start, stop)
                #pylint: disable=len-as-condition
                if lineno < start or len(row) == 0 or row[0].startswith("#"):
                    continue
                elif stop is not None and lineno > stop:
                    break

                if "match" in parser:
                    rex = re.compile(parser["match"]["regex"])
                    col = int(parser["match"]["column"])
                    if not rex.search(row[col]):
                        continue

                row = [item.strip() for item in row]
                result_dict = dict(zip(parser["values"], row))
                yield self.build_result(parser, result_dict)



from collections import OrderedDict
import yaml

class MachinaeLoader(yaml.SafeLoader):
    #pylint: disable=arguments-differ
    def construct_mapping(self, node):
        self.flatten_mapping(node)
        return OrderedDict(self.construct_pairs(node))


MachinaeLoader.add_constructor(
    yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
    MachinaeLoader.construct_mapping)


class MachinaeDumper(yaml.Dumper):
    def represent_dict(self, data):
        return self.represent_mapping('tag:yaml.org,2002:map', data, False)

    def represent_list(self, data):
        return self.represent_sequence('tag:yaml.org,2002:seq', data, False)


MachinaeDumper.add_representer(
    OrderedDict,
    MachinaeDumper.represent_dict)

MachinaeDumper.add_representer(
    list,
    MachinaeDumper.represent_list)

#This is to load site results as an OrderedDict so we override the
#built-in PyYAML safe_load
def safe_load(*args, **kwargs):
    kwargs["Loader"] = MachinaeLoader
    return yaml.load(*args, **kwargs)


def dump(*args, **kwargs):
    kwargs["Dumper"] = MachinaeDumper
    return yaml.dump(*args, **kwargs)


def listsites(conf):
    rstr = '{0:40}{1:40}{2:40}{3}'.format('SITE', 'NAME', 'OTYPES', 'DEFAULT')
    rstr += '\n'
    for key in conf:
        d = 'True'
        if "default" in conf[key].keys():
            d = str(conf[key]["default"])
        rstr += '{0:40}{1:40}{2:40}{3}'.format(key,
                                               conf[key]["name"],
                                               ', '.join(conf[key]["otypes"]),
                                               d)
        rstr += '\n'
    return rstr



from __future__ import absolute_import

import datetime
import json
import re
import urllib.parse
from collections import OrderedDict

from dateutil.parser import parse

from .base import HttpSite


class JsonApi(HttpSite):
    @staticmethod
    def get_value(data, key, default=None):
        if key == "@" or data is None:
            return data
        ret = data
        key_parts = key.split(".")
        for key_part in key_parts:
            if key_part not in ret:
                return default
            ret = ret[key_part]
        return ret

    def get_json(self, url=None):
        r = self.get_content(url=url)

        ignored_status_codes = [int(sc) for sc in self.conf["request"].get("ignored_status_codes", [])]
        if r.status_code in ignored_status_codes:
            return []

        if not self.conf.get("multi_json", False):
            return r.json()

        if r.status_code in ignored_status_codes:
            return []

        results = list()
        for json_line in r.text.split("\n"):
            if not json_line:
                break
            results.append(json.loads(json_line))
        return results

    def run(self):
        data = self.get_json()

        if hasattr(data, "items"):
            next_url = None
            if self.conf.get("paginated", False):
                next_url = data.get("next", None)

            data = [data]
            while next_url:
                next_data = self.get_json(url=next_url)
                next_url = next_data.get("next", None)
                data.append(next_data)

        if "results" not in self.conf:
            return

        for row in data:
            for parser in self.conf["results"]:
                for _ in self.parse_dict(row, parser):
                    yield _

    @classmethod
    def get_result_dicts(cls, data, parser, mm_key=None, onlyif=None):
        if not hasattr(parser, "items"):
            parser = {"key": parser}

        if "key" not in parser:
            yield data
            return

        key = parser["key"]
        rex = None
        if "regex" in parser:
            rex = re.compile(parser["regex"], flags=re.I)

        if key == "@" and mm_key is not None:
            yield {key: mm_key}
            return

        values = cls.get_value(data, key)
        if values is None:
            return

        if not parser.get("match_all", False):
            values = [values]

        for val in values:
            result_dict = OrderedDict()

            if rex:
                m = rex.search(val)
                if not m:
                    return
                #pylint: disable=len-as-condition
                if len(m.groups()) > 0:
                    val = m.groups()
                    if len(val) == 1:
                        val = val[0]

            urldecode = str(parser.get("urldecode", False)).lower()
            if urldecode in ("1", "yes", "true"):
                val = urllib.parse.unquote(val)
            elif urldecode == "twice":
                val = urllib.parse.unquote(
                    urllib.parse.unquote(val)
                )

            if "format" in parser:
                if parser["format"] == "as_list":
                    val = ", ".join(map(str, val))
                elif parser["format"] == "as_time":
                    try:
                        dt = datetime.datetime.fromtimestamp(val)
                    #pylint: disable=bare-except
                    #Will be cleaned up in future refactor -- I hate mcmaster
                    except:
                        dt = parse(val)
                    val = dt.isoformat()
            result_dict[key] = val

            yield result_dict

    @classmethod
    def multi_match_generator(cls, data, parser, mm_key):
        if not hasattr(data, "items"):
            # Is a list, process each list item
            for item in data:
                for _ in cls.multi_match_generator(item, parser, mm_key="@"):
                    yield _

            return

        onlyif = parser.get("onlyif", None)
        if onlyif is not None and not hasattr(onlyif, "items"):
            onlyif = {"key": onlyif}

        # Decide how to iterate on the data
        # Options are:
        #   Return result_dict per match in dict (if: data is dict)
        #   Return one result_dict for whole dict (if: data is dict)
        if mm_key == "@" or parser.get("match_all", False):
            # Treat the entire data as a single match
            # Returns a single result_dict
            data = [(None, data)]
        else:
            # Each matching key is a separate result_dict
            data = data.items()

        for (k, v) in data:
            if onlyif is not None:
                if not hasattr(onlyif, "items"):
                    onlyif = {"key": onlyif}
                value = cls.get_value(v, onlyif["key"], None)

                if value is None:
                    continue
                elif "regex" in onlyif:
                    rex = re.compile(onlyif["regex"], re.I)
                    if not rex.search(value):
                        continue
                else:
                    if not bool(value):
                        continue
            result_dict = OrderedDict()
            for mm_parser in parser["keys"]:
                for mm_result_dict in cls.get_result_dicts(v, mm_parser, mm_key=k, onlyif=onlyif):
                    result_dict.update(mm_result_dict)

            if result_dict:
                result_dict.labels = parser.get("labels", None)
                yield result_dict

    def parse_dict(self, data, parser):
        if not hasattr(parser, "items"):
            parser = {"key": parser}

        if "multi_match" in parser:
            target = self.get_value(data, parser["key"])
            if target is None:
                return
            result_iter = self.multi_match_generator(target, parser["multi_match"], parser["key"])
        else:
            result_iter = self.get_result_dicts(data, parser)

        for result_dict in result_iter:
            yield self.build_result(parser, result_dict)



from ipaddress import ip_address, summarize_address_range

import ipwhois

from .json import JsonApi


class IpWhois(JsonApi):
    @staticmethod
    def get_cidr(network):
        networks = [str(net) for net in summarize_address_range(
            ip_address(network["start_address"]),
            ip_address(network["end_address"])
        )]
        if len(networks) == 1:
            networks = networks[0]
        return networks

    def get_json(self):
        obj = ipwhois.IPWhois(self.kwargs["target"])
        try:
            # import json
            # print(json.dumps(obj.lookup_rdap(depth=2)))
            # return obj.lookup_rdap(depth=2)
            return obj.lookup_rws()
        except AttributeError:
            # rdap = obj.lookup_rdap(inc_raw=True)
            # print(json.dumps(rdap))
            # rdap["network"]["range"] = "{start_address} - {end_address}".format(**rdap["network"])
            # rdap["network"]["cidr"] = self.get_cidr(rdap["network"])
            # return rdap
            # RDAP is a stupid format, use raw whois
            raw = obj.lookup()
            print(raw)
            return raw



from __future__ import absolute_import

import re
from collections import Counter

from bs4 import BeautifulSoup, Comment

from .base import HttpSite


def html_unescape(content):
    try:
        import html
        return html.unescape(content)
    except ImportError:
        import HTMLParser
        html_parser = HTMLParser.HTMLParser()
        return html_parser.unescape(content)


class HtmlSite(HttpSite):
    def get_html(self):
        r = super(HtmlSite, self).get_content()
        body = r.text

        cleanup = self.conf["request"].get("cleanup", {})

        strip_comments = str(cleanup.get("strip_comments", False)).lower()
        if strip_comments in ("1", "yes", "true"):
            soup = BeautifulSoup(r.text, "html5lib")
            for comment in soup.find_all(text=lambda _: isinstance(_, Comment)):
                comment.extract()
            body = str(soup)

        return html_unescape(body)


class TableScraper(HtmlSite):
    @staticmethod
    def compare_rows(row1, row2):
        row1 = [cell.strip().lower() for cell in row1]
        row2 = [cell.strip().lower() for cell in row2]
        #pylint: disable=superfluous-parens
        #I believe these to actually be unnecessary,er,superfluous but will have to test
        return (Counter(row1) == Counter(row2))

    @staticmethod
    def get_row_contents(row):
        return [cell.get_text().strip() for cell in row.find_all(["td", "th"])]

    @classmethod
    def find_table(cls, html, headers):
        soup = BeautifulSoup(html, "html5lib")
        for table in soup.find_all("table"):
            cells = cls.get_row_contents(table.find("tr"))
            if cls.compare_rows(cells, headers):
                return (table, cells)
        raise ValueError("No matching table found")

    def run(self):
        body = self.get_html()

        for parser in self.conf["results"]:
            (table, columns) = self.find_table(body, parser["map"].keys())
            for row in table.find_all("tr"):
                cells = self.get_row_contents(row)
                if self.compare_rows(cells, columns):
                    continue
                result_dict = dict(zip(columns, cells))
                yield self.build_result(parser, result_dict)


class Webscraper(HtmlSite):
    def run(self):
        body = self.get_html()

        if "results" not in self.conf:
            raise Exception("No parsing configuration found")
        for parser in self.conf["results"]:
            rex = re.compile(parser["regex"], flags=re.I)
            for match in rex.finditer(body):
                result_dict = dict()
                for (k, v) in zip(parser["values"], match.groups()):
                    result_dict[k] = v
                yield self.build_result(parser, result_dict)



import subprocess
from setuptools import setup, find_packages
from os import path

#this should hopefully allow us to have a more pypi friendly, always up to date readme
readMeDir = path.abspath(path.dirname(__file__))
with open(path.join(readMeDir, 'README.md'), encoding='utf-8') as readFile:
    long_desc = readFile.read()


VERSION = '1.4.5'

setup(
    name='machinae',
    version=VERSION,
    author='Steve McMaster',
    author_email='mcmaster@hurricanelabs.com',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,
    zip_safe=False,
    url='http://hurricanelabs.github.io/machinae/',
    description='Machinae Security Intelligence Collector',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    install_requires=[
        'dnspython3',
        'ipwhois<0.11',
        'requests',
        'stopit',
        'pyyaml',
        'beautifulsoup4',
        'html5lib',
        'relatime',
        'tzlocal',
        'filemagic',
        'feedparser',
        'defang',
    ],
    entry_points={
        'console_scripts': [
            'machinae = machinae.cmd:main',
        ]
    },
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Development Status :: 5 - Production/Stable',
    ],
    bugtrack_url='https://github.com/HurricaneLabs/machinae/issues',
)



import collections
import re
import socket
import ipaddress


TargetInfo = collections.namedtuple("TargetInfo", ("target", "otype", "otype_detected"))
ErrorResult = collections.namedtuple("ErrorResult", ("target_info", "site_info", "error_info"))
ResultSet = collections.namedtuple("ResultSet", ("target_info", "results"))
SiteResults = collections.namedtuple("SiteResults", ("site_info", "resultset"))
Result = collections.namedtuple("Result", ("value", "pretty_name"))

#pylint: disable=no-else-return,too-many-return-statements
def get_target_type(target):
    try:
        getVer = ipaddress.ip_address(target)
        if getVer.version == 4:
            return "ipv4"
        elif getVer.version == 6:
            return "ipv6"
    except ValueError:
        pass

    #pylint: disable=no-else-return
    # Hashes
    if re.match("^[a-f0-9]{32}$", target, re.I):
        # MD5
        return "hash"
    elif re.match("^[a-f0-9]{40}$", target, re.I):
        # SHA-1
        return "hash.sha1"
    elif re.match("^[a-f0-9]{64}$", target, re.I):
        # SHA-256
        return "hash.sha256"
    elif re.match("^[a-f0-9]{128}$", target, re.I):
        # SHA-512
        return "hash.sha512"

    # URL
    elif re.match("^https?://", target, re.I):
        return "url"

    # Email Addresses
    elif re.match("^.*?@.*?$", target, re.I):
        return "email"

    # SSL fingerprints
    elif re.match("^(?:[a-f0-9]{2}:){19}[a-f0-9]{2}$", target, flags=re.I):
        return "sslfp"

    return "fqdn"


# d2 takes precedence
def dict_merge(d1, d2):
    d3 = d1.copy()
    for key in d2:
        if key in d3 and hasattr(d3[key], "items") and hasattr(d2[key], "items"):
            d3[key] = dict_merge(d3[key], d2[key])
        elif hasattr(d2[key], "items"):
            d3[key] = d2[key].copy()
        else:
            d3[key] = d2[key]
    return d3
